/*
 * Checkpoint.c
 *
 *  Created on: Mar 2, 2013
 *      Author: Brenda Javornik (e-mail: brendaj (at) cora.nwra.com)
 */

#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <err.h>
#include <unistd.h>
#include <stdio.h>
#include <stdbool.h>
#include <stdlib.h>
#include <string.h>
#include "errorHandler.h"
#include "utilities.h"
#include "BaseDirectory.h"
#include "MapDepthMatrix.h"
#include "OLABaseErrorMessages.h"


const static char CheckpointConstantCkptDir[] = "ckpt";
const static char CheckpointConstantCkptFile0[] = "w_ckpt0.fits";
const static char CheckpointConstantCkptFile1[] = "w_ckpt1.fits";

/**
 * Structure holding checkpoint information.
 */
struct CheckpointData {
	 long Value;  			// Dump checkpoints after this many iterations of lsqr
	 char *Filename;		// Full path to previous checkpoint file to use as starting
	 	 	 	 	 	 	// point. This overrides the initial guess file in the input
	 	 	 	 	 	 	// parameter file.
	 char *ckptDirName;		// Full path to the checkpoint directory
	 char *ckptFileName0; 	// Full path to the checkpoint file names
	 char *ckptFileName1;
};

/**
 * Initial values for the checkpoint data.
 */
static struct CheckpointData CheckpointInfo = {
		0,
		NULL,
		NULL,
		NULL,
		NULL
};

long getCheckpoint() {
	return CheckpointInfo.Value;
}

bool runningWithCheckpoints() {
	return (getCheckpoint() > 0L);
}

/**
 * Remove all memory associated with checkpoint file and path names.
 */
void destroyCheckpoint() {

	if (runningWithCheckpoints()) {
		if (CheckpointInfo.Filename != NULL)
			td_free(CheckpointInfo.Filename, strlen(CheckpointInfo.Filename) + 1);
		td_free(CheckpointInfo.ckptFileName0, strlen(CheckpointInfo.ckptFileName0) + 1);
		td_free(CheckpointInfo.ckptFileName1, strlen(CheckpointInfo.ckptFileName1) + 1);
		td_free(CheckpointInfo.ckptDirName, strlen(CheckpointInfo.ckptDirName) + 1);
	}
}

/**
 * Remove the checkpoint sub-directory
 *
 */
int removeCheckpoints() {

	if (runningWithCheckpoints()) {
		// remove the checkpoint files because the inversion finished successfully
		//int remove(const char *pathname);
		// On success, zero is returned.  On error, -1 is returned, and errno is set appropriately.
		int result;
		result = remove(CheckpointInfo.ckptFileName0);
		if (result != 0) {
			int errsv = errno;
	    	reportErrorNo(errsv, CheckpointInfo.ckptFileName0, "removeCheckpoints", WARNING);
		}
		result = remove(CheckpointInfo.ckptFileName1);
		if (result != 0) {
			int errsv = errno;
	    	reportErrorNo(errsv, CheckpointInfo.ckptFileName1, "removeCheckpoints", WARNING);
		}
		result = remove(CheckpointInfo.ckptDirName);
		if (result != 0) {
			int errsv = errno;
	    	reportErrorNo(errsv, CheckpointInfo.ckptDirName, "removeCheckpoints", WARNING);
		}

	}
	return 0;
}

/**
 * Rename the checkpoint sub-directory
 *
 */
int renameCheckpointFile() {
	int status;

	printf("renaming checkpoint directory\n");

	char *parent, *child, *baseDir;

	getParentNChild(CheckpointInfo.Filename, &parent, &child);
	// free space for child, since we don't need it
	td_free(child, strlen(child) + 1);
	getParentNChild(parent, &baseDir, &child);
	char *OldName = constructFileName(baseDir, "ckptOld");
	printf("OldName = %s\n", OldName);

	// On success, zero is returned.  On error, -1 is returned, and errno is set appropriately.
	status = rename(parent, OldName);
	if (status != 0) {
		int errsv = errno;
		reportErrorNo(errsv, CheckpointInfo.ckptDirName, "renameCheckpointFile",
				FATAL);
	}
	td_free(OldName, strlen(OldName) + 1);
	td_free(child, strlen(child) + 1);
	td_free(baseDir, strlen(baseDir) + 1);
	td_free(parent, strlen(parent) + 1);
	return status;
}

/**
 * Record the path and name of a previous checkpoint file.
 * @param[in] filename Full path to a previous checkpoint file.
 */
void setCheckpointInputFile(char *filename) {
	if (CheckpointInfo.Filename == NULL) {
		CheckpointInfo.Filename = filename;
		reportErrorStr("Setting input checkpoint file to", CheckpointInfo.Filename,
				"setCheckpointInputFile", INFO);
	} else {
		reportErrorStr("found more than one checkpoint file, which is confusing",
				CheckpointInfo.Filename, "setCheckpointInputFile", FATAL);
	}
}


/**
 * Return the full path to the latest checkpoint file.
 */
char *getCheckpointInputFile() {
	return CheckpointInfo.Filename;
}

/**
 * Look for checkpoint files in the given directory.
 *  If any are found, record the full path and file name
 *  for use later.
 *  @param[in] dirName Directory in which to look for
 *  specially named checkpoint files.  Find the most
 *  recent one.
 */
void lookForLatestCkptFile(char *dirName) {

	struct dirent *dp;
	DIR *dirPtr;

	if ((dirPtr = opendir(dirName)) == NULL) {
        int errsv = errno;
    	reportErrorNo(errsv, dirName, "lookForLatestCkptFile", FATAL);
	}
	time_t prevCkptTime = 0;

	char *latestCkpt = NULL;
	while ((dp = readdir(dirPtr)) != NULL) {
		if (!strcmp(dp->d_name, ".") || !strcmp(dp->d_name, ".."))
			continue; // skip self and parent
		if ((strcmp(dp->d_name, CheckpointConstantCkptFile0) == 0)
				|| (strcmp(dp->d_name, CheckpointConstantCkptFile1) == 0)) {
			// ok, we've found a checkpoint file
			// construct the full path to it and check its status
			char *name;
			name = constructFileName(dirName, dp->d_name);

			// check the last modify time; if it's later than the previous one, switch to this one.
			struct stat stbuf;

			int result;
			result = stat(name, &stbuf);
			if (result < 0) {
				reportError(OLABaseNoStatCkpt,
						"lookForLatestCkptFile", WARNING);
			} else {
				if ((stbuf.st_mtime - prevCkptTime) > 0) {
					prevCkptTime = stbuf.st_mtime;
					if (latestCkpt != NULL) {
						td_free(latestCkpt, strlen(latestCkpt) + 1);
					}
					latestCkpt = td_malloc(strlen(name) + 1);
					strcpy(latestCkpt, name);
				}
			}
			td_free(name, strlen(name) + 1);
		} // end if (name  == "ckpt0.fits")
	} // end while (dp = readdir)

	closedir(dirPtr);
	if (latestCkpt != NULL) {
		setCheckpointInputFile(latestCkpt);
	}
}

/**
 * Look for checkpoint directory
 * @param[in] dirPtr2 Pointer to starting directory structure
 * @param[in] dirName Name of starting directory for search.
 */
void lookForCkptDir(DIR *dirPtr2, char *dirName) {
	DIR *dirPtr;

    struct dirent **namelist;
    int n, N;

    // internal function that returns true if the subdirectory
    // matches the checkpoint directory name
	int filter(const struct dirent *dp) {
		if (strcmp(dp->d_name, CheckpointConstantCkptDir) != 0)
			return 0;
		else
			return 1;  // we are done, it matches
	}

	if ((dirPtr = opendir(dirName)) == NULL) {
        int errsv = errno;
    	reportErrorNo(errsv, dirName, "lookForCkptDir", FATAL);
	}

	n = scandir(dirName, &namelist, filter, alphasort);
	N = n;
	if (n < 0) {
		reportError(OLABaseNoDirCkpt, "lookForCkptDir", INFO);
	}

	n = N;
	while (n--) {
		char *myname = constructFileName(dirName, namelist[n]->d_name);
		lookForLatestCkptFile(myname);
		td_free(myname, strlen(myname) + 1);
		free(namelist[n]);
	}
	free(namelist);
}

/** Setup everything for checkpoints.  Find existing checkpoint directory for restarting,
 * setup the checkpoint filenames.
 * @param[in] value Write a checkpoint file after this many iterations of lsqr.
 * @param[in] outputBaseDir Be sure to send the original output base directory, not the one
 * that has version(s).  This is where we'll look for an existing checkpoint directory.
 * @param[in] versionedOutputBaseDir Path for output base directory.  This is where the
 * checkpoint files will go.
 */
void setCheckpoint(long value, char *outputBaseDir, char *versionedOutputBaseDir) {

    char *parent;
    char *child;

	int filter(const struct dirent *dp) {

		// Need to have this function within a function that knows about the
		// output base directory in order to determine the length of the
		// string without passing either the length or the directory name
		// to the filter function.
		int len;
		len = strlen(child);

		if (strncmp(dp->d_name, child, len) != 0)
			return 0;
		// what can come next? end-of-string, or _v9999
		if (len == strlen(dp->d_name))
			return 1;  // we are done, it matches
		// now look for _vnnnn, where n is in [0-9]
		if (strncmp(&(dp->d_name[len]), "_v", 2) != 0)
			return 0;
		// start at first digit of version number
		int p;
		p = len + 2 + 1;
		// loop while character is in [0-9]
		// if dp->d_name[i] >= '0' and <= '9', then ok
		// ok if reached end of string, otherwise, discard this name
		bool stillOk;
		stillOk = true;
		while ((dp->d_name[p] != '\0') && stillOk) {
			if ((dp->d_name[p] <'0') || (dp->d_name[p] > '9'))
				stillOk = false;
			p++;
		}
		if (stillOk)
			return 1;
		else
			return 0;
	}

	CheckpointInfo.Value = value;

	// if we are running with checkpoints, then look for checkpoint files
	if (value > 0L) {

        struct dirent **namelist;
        int n, N;

        // get the parent of the outputBaseDir
        // check for root (/), wait, output basename cannot be root! check for this!
        // check for just one name (name), which would mean (.)
        // can I just access the ".." from the readdir? nope.

        getParentNChild(outputBaseDir, &parent, &child);

        if (parent == NULL)
        	reportError("Could not determine parent directory of output base directory, exiting", "setCheckpoint", FATAL);

        // look for <output-base-dir><_v*>/ckpt directory

        DIR *dirPtr;
        dirPtr = exists(parent);
        if (dirPtr != NULL) {
			n = scandir(parent, &namelist, filter, alphasort);
			N = n;
			if (n < 0)
				perror("scandir");
			else {
				while (n--) {
					printf("%s\n", namelist[n]->d_name);
				}
			}

			n = N;
			while (n--) {
				DIR *cdirPtr;
				cdirPtr = getDirectory(namelist[n]->d_name);
				char *myname = constructFileName(parent, namelist[n]->d_name);
				lookForCkptDir(cdirPtr, myname);
				td_free(myname, strlen(myname) + 1);
			}
			char *ckptFile;
			ckptFile = getCheckpointInputFile();
			if (ckptFile != NULL)
				reportErrorStr("Latest checkpoint file is", ckptFile, "setCheckpoint", INFO);
			else {
				reportError("No checkpoint files found", "setCheckpoint", INFO);
			}
		} else {
        	// base directory doesn't exist, there are no ckpt files to access
        	reportError("No existing checkpoint files found", "setCheckpoint", INFO);
        }
	}
	td_free(parent, strlen(parent) + 1);
	td_free(child, strlen(child) + 1);

	// Ok, the checkpoint files are always going to be named one of two values,
	// so let's just set the names now and then reuse them.
	//CheckpointInfo.OutputBaseDir = newBaseDirectory(versionedOutputBaseDir);
	// char *constructFileName(char *baseDir, char *fileName)
	CheckpointInfo.ckptDirName = constructFileName(versionedOutputBaseDir, CheckpointConstantCkptDir);
	CheckpointInfo.ckptFileName0 = constructFileName(CheckpointInfo.ckptDirName, CheckpointConstantCkptFile0);
	CheckpointInfo.ckptFileName1 = constructFileName(CheckpointInfo.ckptDirName, CheckpointConstantCkptFile1);
}


/**
 * Write OLA inversion weights, w, to a checkpoint file in a sub-directory.
 * @param[in] w MAP_DEPTH_MATRIX containing the solution weights.
 * @param[in] parms FITSParameters contains header parameters and values for FITS file.
 * @param[in] baseName BaseDirectory structure for managing output file name.
 */
void writeCheckpoint(MAP_DEPTH_MATRIX w, FITSParameters parms, BaseDirectory *baseName) {
	bool useFT = true;
	char *filename;
	static bool firstCheckpoint = true;
	static bool firstTime = true;

	// I think checkpoint should hold it's own version of the output base directory
	// It should be dynamically allocated (and freed) of length outputBaseDir + strlen(/ckpt/w_ckpt0.fits)
	if (firstTime) {
		// make sure the sub-directory for checkpoint files is created
		if (getDirectory(CheckpointInfo.ckptDirName) == NULL)
			reportError("Cannot create checkpoint directory", "writeCheckpoint", FATAL);
		firstTime = false;
	}

	// we want something like this: <basedir>/ckpt/w_ckpt0.fits
	if (firstCheckpoint) {
		filename = CheckpointInfo.ckptFileName0;
	} else {
		filename = CheckpointInfo.ckptFileName1;
	}
	firstCheckpoint = !firstCheckpoint;

	printf("writing checkpoint to %s\n", filename);
	writeMapDepthMatrixToOneFile(filename, w, useFT, parms);
}

/**
 * Create sub-directory inside output base directory.  The sub-directory
 * contains the checkpoint files.
 */
bool creatCheckpointSubDir() {
	bool ok;
	ok = true;

	return ok;
}
