/*
 * Parameter.c
 *
 *  Created on: Apr 30, 2012
 *      Author: Brenda Javornik (e-mail: brendaj (at) nwra.com)
 */

#include <stdio.h>
#include "Parameter.h"
#include "lambda.h"
#include "Target.h"
#include "CrossTalk.h"
#include "WeightSpread.h"


// Global variables
// Parameters is a pointer to the head of a linked list of parameters.
// Mu, Nu, and Sigma are pointers to individual elements/nodes of the
// parameter list.

static ParameterType *Parameters = NULL;
static ParameterType *Nu = NULL;
static ParameterType *Mu = NULL;
static ParameterType *Sigma = NULL;
static ParameterType *Epsilon = NULL;

/**
 * Get the list of looping parameters.
 */
ParameterType *getParameters() {
	return Parameters;
}

/**
 * Get a pointer to looping parameter, Mu.
 */
ParameterType *getMu() {
	return Mu;
}

void setMu(ParameterType *pt) {
	Mu = pt;
}

/**
 * Get a pointer to looping parameter, Nu.
 */
ParameterType *getNu() {
	return Nu;
}

void setNu(ParameterType *pt) {
	Nu = pt;
}

/**
 * Get a pointer to looping parameter, Epsilon.
 */
ParameterType *getEpsilon() {
	return Epsilon;
}

void setEpsilon(ParameterType *pt) {
	Epsilon = pt;
}

/**
 * Get a pointer to looping parameter, Sigma.
 */
ParameterType *getSigma() {
	return Sigma;
}

void setSigma(ParameterType *pt) {
	Sigma = pt;
}

void setSigmaStep(int steps) {
	Sigma->parameterRange.n = 1;
}

/**
 * Increment parameter to next value in its range.
 * @param[in] pt ParameterType Pointer to a looping parameter, for example, mu, sigma, lamda, nu.
 */
void incrementParameter(ParameterType *pt) {
	if (pt != NULL)

	// bump parameter to the next step
	pt->currentStep += 1;
	// if parameter exceeds max, then reset to minimum value and bump sigma
	if (pt->currentStep > pt->parameterRange.n) {
		// bump next parameters to next step
		ParameterType *dependantPt;
		dependantPt = pt->next;
		if (dependantPt != NULL) {
			dependantPt->increment(dependantPt);
			pt->currentStep = 1;
			pt->currentValue = pt->parameterRange.min;
		}
	} else {
		// get the value of parameter for this step
		pt->currentValue =
				nextLambdaValue(pt->parameterRange, pt->currentStep-1);
	}
}

/**
 * Return true, if there are more values in the parameter range.
 * @param[in] pt ParameterType Pointer to a looping parameter.
 * @return True if more values, false if the all values in range
 * have been used.
 */
bool moreParameters(ParameterType *pt) {
	if (pt == NULL)
		return true;
	else {
		if (pt->currentStep <= pt->parameterRange.n)
			return(pt->more(pt->next));
		else
			return false;
	}
}

/**
 * Add a LambdaSet that specifies a parameters range to the list of
 * looping parameters.
 * @param[in] ls LambdaSet (dependency Pointer to ParameterData?? not sure what this means.)
 */
void addParameter(LambdaSet ls) {
	// always add to the front of the list
	// access global variable Parameters
	ParameterType *pd;
	pd = (ParameterType *) td_malloc(sizeof(ParameterType));

	pd->parameterRange = ls;

	pd->currentValue = ls.min;
	pd->currentStep = 1;
	pd->increment = &incrementParameter;
    pd->more = &moreParameters;

	// make new parameter the head of the list
	pd->next = Parameters;
	Parameters = pd;
}

/** Special function to increment sigma, since it requires computing
 * a new target function.
 * @param[in] pt  A pointer to a ParameterType.
 */
void incrementSigma(ParameterType *pt) {
	if (pt != NULL) {
		// bump parameter to the next step
		pt->currentStep += 1;
		// if parameter exceeds max, then reset to minimum value and bump sigma
		if (pt->currentStep > pt->parameterRange.n) {
			// bump next parameters to next step
			ParameterType *dependantPt;
			dependantPt = pt->next;
			if (dependantPt != NULL) {
				dependantPt->increment(dependantPt);
				pt->currentStep = 1;
				pt->currentValue = pt->parameterRange.min;
			}
		} else {
			// get the value of parameter for this step
			pt->currentValue =
					nextLambdaValue(pt->parameterRange, pt->currentStep-1);
			// if next sigma is within range, get a new target function
			resetTarget(pt->currentValue, pt->currentStep);
			printf("reset target\n");
		}
	}
}

/**
 * Initialize the looping parameters using information from the
 * input file.
 * @param[in] inputParms InputParms structure containing parameter
 * values parsed from the input file.
 */
void initLoopParameters(InputParms inputParms) {
	addParameter(inputParms.lambdaSet2);
	setSigma(getParameters());
	getSigma()->increment = &incrementSigma;
	addParameter(inputParms.lambdaSet1);
	setMu(getParameters());
	if (inputParms.xt) {
		addParameter(inputParms.lambdaSet3);
		setNu(getParameters());
		setCrossTalk();
	} else {
	    LambdaSet *lsPtr;
	    // set a default value for nu (equal to zero), if not using crosstalk parameter.
	    // Also related to Model.setParameters2() when nValues = 2, nu is set to default value.
	    // newLambdaSet(min, max, linearSteps, logSteps)
		lsPtr = newLambdaSet(0, 1, 1, 0);
		addParameter(*lsPtr);
		freeLambdaSet(lsPtr);
		setNu(getParameters());
	}
	if (inputParms.spread) {
		addParameter(inputParms.lambdaSet4);
		setEpsilon(getParameters());
		if (inputParms.constrained) {
			setWeightSpreadCLSQR();
		} else {
			setWeightSpread();
		}
	} else {
	    LambdaSet *lsPtr;
	    // set a default value for epsilon (equal to zero), if not using spread parameter.
	    // Also related to Model.setParameters2() when nValues = 2, nu is set to default value.
	    // newLambdaSet(min, max, linearSteps, logSteps)
		lsPtr = newLambdaSet(0, 1, 1, 0);
		addParameter(*lsPtr);
		freeLambdaSet(lsPtr);
		setEpsilon(getParameters());
	}
}

/**
 * Free all memory associated with a ParameterType.
 * @param[in] pt  A pointer to a ParameterType.
 */
void destroyParameters(ParameterType *pt) {
	ParameterType *next;

	while (pt != NULL) {
		next = pt->next;
		td_free(pt, sizeof(ParameterType));
		pt = next;
	}
}

/**
 * send already allocated space for parmValues and parmSteps
 */
void doSomething(Float *parmValues, int *parmSteps, int *nParms) {

	int sigmaStep, muStep, nuStep, epsilonStep;
	Float mu, sigmah, nu, epsilon;

	sigmaStep = getSigma()->currentStep;
	muStep = getMu()->currentStep;
	nuStep = getNu()->currentStep;
	epsilonStep = getEpsilon()->currentStep;
	mu = getMu()->currentValue;
	sigmah = getSigma()->currentValue;
	nu = getNu()->currentValue;
	epsilon = getEpsilon()->currentValue;

	// we always need mu and sigma; crosstalk (nu) is optional
	*nParms = 2;
	*parmSteps++ = muStep;
	*parmSteps++ = sigmaStep;
	*parmValues++ = mu;
	*parmValues++ = sigmah;
	printf("\tmu = %10.3e; \tsigma_h = %10.3e", mu, sigmah);
	if (getCrossTalk()) {
		*nParms = 3;
		*parmSteps++ = nuStep;
		*parmValues++ = nu;
		printf(";\tnu = %10.3e", nu);
	}
	if (getWeightSpread() || getWeightSpreadCLSQR()) {
		*nParms = 4;
		*parmSteps = epsilonStep;
		*parmValues = epsilon;
		printf(";\tepsilon = %10.3e", epsilon);
	}
	printf("\n");
}

