/*
 * Target.c
 *
 *  Created on: Aug 16, 2010
 *      Author: Brenda Javornik (e-mail: brendaj (at) nwra.com)
 */

#include <math.h>
#include <complex.h>
#include <cblas.h>
#include "errorHandler.h"
#include "OLABaseErrorMessages.h"
#include "objectfits.h"
#include "matrix.h"
#include "fftw.h"
#include "MapDepthMatrix.h"
#include "ProblemDimension.h"
#include "Gaussian.h"
#include "Scatterer.h"
#include "Constants.h"
#include "BaseDirectory.h"
#include "objectfits.h"
#include "Parameter.h"

#define NORM 1 // switch to choose different normalization schemes for Target function.

#ifdef SINGLE
#define cblas_dasum cblas_sasum
#define cblas_dscal	cblas_sscal
#define cblas_zcopy cblas_ccopy
#else
#define fftwf_free  fftw_free
#endif

/**
 * @defgroup Target Target function - The target function for OLA Averaging Kernels.
 * The target function is implemented as a 2-D Gaussian in x and y, multiplied by a a 1-D function of z.
 * The 1-D function is an input parameter.
 * The Gaussian has an input parameter, sigma_h, which specifies the horizontal
 * full-width at half-maximum (FWHM) of the function.
 *
 * The target function is precalculated for x, y and z values, then Fourier transformed.
 *
 */


// This can be a 2D matrix (if there are multiple target depths), instead of 1xNz
// for a single target depth.
///> Single instance of input parameter f(z; z0)
static MATRIX Target_f;

///> Single instance of the target
static MAP_DEPTH_MATRIX *Target = NULL;

struct TargetData {
	int N;
	int Ny;
	int Nz;
	int Nfz0;
	bool useFFT;					///< True if using Fourier Transform; False otherwise.
	bool debug;
	BaseDirectory *outputBaseDir;
};

static struct TargetData targetData;

void my_zset(const int N, void *alpha, void *X, const int incX) {
	// Nz, &complexZero, ZeroStartPtr, 1);
	int i;
	FloatComplex *ptr;
	ptr = X;
	for (i=0; i<N; i++) {
		memcpy(ptr, alpha, sizeof(FloatComplex));
		ptr += incX;
	}
}

/**
 * Verify the dimensions of the target function are compatible with the dimensions of the kernels.
 * Oh, but if the inversion is NOT OLA, then only allow (1xNz) ... so don't use this function.
 * @param[in,out] f 1D function f(z) for target function when z > 1.  This parameter is in/out
 * 					because it is NFz0 x Nz, and if the FITS file is Nz x NFz0, then we can easily take
 * 					the transpose and still run the inversions.
 * @param[in] Nz Size of the z-dimension of the kernels.
 * @returns bool Returns true if dimensions are compatible, false otherwise.
 * @ingroup Target
 */
bool checkTargetDim(MATRIX *f, int Nz) {
	bool ok = true;
	char msg2[MAX_ERROR_MSG_LENGTH];

	// we want the number of rows to be the number of depths
	// and the number of columns to be Nz.  So, the function spans a row.

	// Make sure f.nRow or f.nCol = K.nZ; orient as needed with transpose; f needs to be Zx1
	if (f->nCol != Nz) {
		// try transpose
		if (f->nRow == Nz) {
			// f is ok, just take the transpose
			printf("taking transpose of target depth function\n");
			MATRIX fT = transpose(*f);
			freeMatrix(*f);
			*f = fT;
		} else {
			ok = false;
			// error, "f needs to have dimensions %d x 1", Nz
			snprintf(msg2, MAX_ERROR_MSG_LENGTH, "%s %d x Nfz0, where Nfz0 >= 1", OLABaseTargetfDim, Nz);
			reportError(msg2, "main", ERROR);
		}
	}

	if (ok)
		printf("Target depth function (fz0) has dimensions: Nz=%d, solving for %d target depths\n",
				f->nCol, f->nRow);
	return ok;
}

/**
 * Set new parameters for Target function.
 * @param[in] f A MATRIX that contains a 1D function for z-dimension;
 * but it can contain multiple 1D functions, one for each target depth.
 * @param[in] N X-dimension of target function in frequency space.
 * @param[in] Ny Y-dimension in frequency space if useFourierTransform  is true;
 *  different than Nx because of Fourier Transform; otherwise, Y-dimension in real space.
 * @param[in] Nz Z-dimension.
 * @param[in] baseDir Output base directory.
 * @param[in] debug True if running with debug flag; False otherwise.
 * @param[in] useFourierTransform True if transforming target function to wavespace; False otherwise.
 * Multiply 2D Gaussian by this function to get 3D Target function.
 * @ingroup Target
 */
void newTarget(MATRIX f, int N, int Ny, int Nz, BaseDirectory *baseDir, bool debug,
		bool useFourierTransform) {
// send new value MATRIX f(z,z0) = 1D function
	Target_f = f;
	if (debug) print_matrix2("Target_f", Target_f, ROW_MAJOR);
	targetData.N = N;
	targetData.Ny = Ny;
	targetData.Nz = Nz;
	targetData.Nfz0 = f.nRow; // store all values for a depth function as a row
	targetData.useFFT = useFourierTransform;
	targetData.debug = debug;
	targetData.outputBaseDir = baseDir;
}

void NormalizeFz() {

	MATRIX fz;
	fz = Target_f;

	// normalize fz, the depth vector, for each depth, Nfz0
	Float *fp;
	fp = fz.data;
	int nfz;
	for (nfz = 0; nfz<fz.nRow; nfz++) {
		Float sumOfFi = 0.0;

		// scale Target_f by 1/sumOfFi, so that the spatial integral of the Target function is one.
		int incX = 1;
		sumOfFi =  cblas_dasum  (fz.nCol, fp, incX);
		if (sumOfFi == 0.0) {
			reportError("target function is zero", "constructTarget", FATAL);
		}
		Float alpha;
		alpha = 1.0/sumOfFi;

		cblas_dscal(fz.nCol, alpha, fp, incX);

		fp += fz.nCol; // move to the next row
	}

}

void NormalizeFzLSQR() {

	Float *fp;
	fp = Target_f.data;
	Float sumOfFi = 0.0;

	int i;
	for (i=0; i<Target_f.nCol*Target_f.nRow; i++)
		sumOfFi += fp[i];
	// scale Target_f by 1/sumOfFi, so that the spatial integral of the Target function is one.
	if (sumOfFi == 0.0) {
		reportError("target function is zero", "constructTarget", FATAL);
	}
	MATRIX temp = multiplyScalar(Target_f, 1.0/sumOfFi);

	freeMatrix(Target_f);
	Target_f = temp;
}

/*
 * Really, this should be called scaleDepthByKxKy because we are scaling the
 * target depth function by each 2DTarget(kx,ky) element
 * @param[in] ix Index of kx,ky element in 2D target function.
 * @param[in] iy Index of kx,ky element in 2D target function.
 * @param[in,out] T Pointer to MAP_DEPTH_MATRIX containing target function.
 * @param[in] scaleFactor Scalar value at 2DTarget(kx,ky)
 * @param[in,out] tp Pointer inside Target where changes are made.
 */
void Buildup(int ix, int iy, MAP_DEPTH_MATRIX *T, FloatComplex *scaleFactor, FloatComplex **tp) {

	MATRIX fz;
	fz = Target_f;
	FloatComplex *fzp;

	fzp = fz.data;
	// scale the depth functions by the Gaussian xy-plane elements
	// for each depth
	int d;
	for (d = 0; d < targetData.Nfz0; d++) {
		// cblas_zcopy (N, const void *X, const int incX, void *Y, const int incY)
		cblas_zcopy(fz.nCol, fzp, 1, *tp, 1);
		cblas_zscal(fz.nCol, scaleFactor, *tp, 1);

		*tp += T->nZ * T->nV;  // move to same vx,vy, or vz, in next row of Target
		// which corresponds to the next fz0
		fzp += fz.nCol; // move to the next row
	} // end for each depth
}

/**
 *
 * Really, this should be called scaleDepthByKxKy because we are scaling the
 * target depth function by each 2DTarget(kx,ky) element.
 * Ok, the real reason for two versions of this function is
 * because multiplyScalarZ produces slightly different numbers than
 * cblas_zscal, and this causes all the results of tests to be
 * slightly different, and I don't want to change all the expected
 * output files for testing because they are huge.
 * @param[in] ix Index of kx,ky element in 2D target function.
 * @param[in] iy Index of kx,ky element in 2D target function.
 * @param[in,out] T Pointer to MAP_DEPTH_MATRIX containing target function.
 * @param[in] scaleFactor Scalar value at 2DTarget(kx,ky)
 * @param[in,out] tp Pointer inside Target where changes are made.
 */
void BuildupLSQR(int ix, int iy, MAP_DEPTH_MATRIX *T, FloatComplex *scaleFactor,
		FloatComplex **tp) {

	MATRIX Tfz = multiplyScalarZ(Target_f, *scaleFactor);

	FloatComplex *fzp;
	fzp = Tfz.data;
	// scale the depth functions by the Gaussian xy-plane elements
	// cblas_zcopy (N, const void *X, const int incX, void *Y, const int incY)
	cblas_zcopy(Target_f.nCol, fzp, 1, *tp, 1);
	freeMatrix(Tfz);
	*tp += T->nZ * T->nV;  // move to same vx,vy, or vz, in next row of Target
}

/**
 * Prints target function to the screen.
 * @param[in] T A MAP_DEPTH_MATRIX containing the target function.
 */
void printTarget(MAP_DEPTH_MATRIX T) {
	int kx,ky;
	for (kx=0; kx<T.nX; kx++) {
		for (ky=0; ky<T.nY; ky++) {
			MATRIX md;
			md = extractMapDepth(kx,ky,T);
			printf("kx,ky = %d, %d\n", kx,ky);
			print_matrix2("md", md, ROW_MAJOR);
		}
	}
	printf("------------\n");
}

/**
 * Construct target for multiple depths
 *
 * Construct Target function from preset parameters (Gaussian keeps sigma_h,
 * Target_f should already be set to f(z;z0) using newTarget() function).
 * @return Target function stored in a 4-D MAP_DEPTH_MATRIX in wavespace.
 * @ingroup Target
 */
MAP_DEPTH_MATRIX *constructTarget(
		void (*NormalizeFz) (),
		void (*Buildup) (int, int, MAP_DEPTH_MATRIX *, FloatComplex *, FloatComplex **)
		) {
	Float value;
	int ix, iy; // indices into matrices
	Float x, y; // values for calculating target function
	// because center point is at x = 0, y = 0, z = 0
	// which corresponds to ix=Nx/2, etc.
	MATRIX t;
	bool shift = false;
	static MAP_DEPTH_MATRIX T;
	int dim;
	int Nv;
	int Nx, Ny, Nz, Nfz0;
	bool debug;

	Nx = targetData.N;
	Ny = targetData.Ny;
	Nz = targetData.Nz;
	Nfz0 = targetData.Nfz0;

	debug = targetData.debug;

	dim = getProblemDimension();
	if (dim == 5)
		Nv = 3;
	else
		Nv = 1;

	static int DataStartOffset;

	// Do only once!!!
	// create holding matrix in frequency space coordinates
	if (T.data == NULL ) {
		Float *Scatterer = getScatterer();

		T = newMapDepthMatrixV(Nx, Ny, Nz, Nfz0, Nv, shift);

		if (dim == 5) {
			// determine where to start target data; where to start zeros;
			// how long are zeros; how many zero sections
			// I can zero all sections at once and only once, because they don't change
			// only the target data change for each sigma_h !!!
			FloatComplex complexZero = 0.0 + 0.0 * I;
			FloatComplex *ZeroStartPtr;
			if (Scatterer[0] == 1) {
				DataStartOffset = 0;
				ZeroStartPtr = T.data + Nz;
			} else if (Scatterer[1] == 1) {
				DataStartOffset = Nz;
				ZeroStartPtr = T.data;
				// zero first segment, then the rest are regularly spaced
				my_zset(Nz, &complexZero, ZeroStartPtr, 1);
				ZeroStartPtr = T.data + Nz * 2;
				int i;
				printf("Nx=%d, Ny=%d, Nfz0=%d, Nz=%d\n", Nx, Ny, Nfz0, Nz);
				for (i = 0; i < Nx * Ny * Nfz0 - 1; i++) {
					my_zset(Nz * 2, &complexZero, ZeroStartPtr, 1);
					ZeroStartPtr += Nz * Nv;
				}
				// zero last segment
				my_zset(Nz, &complexZero, ZeroStartPtr, 1);
			} else if (Scatterer[2] == 1) {
				ZeroStartPtr = T.data;
				DataStartOffset = Nz * 2;
			} else {
				reportError("Scatterer is not set correctly", "constructTarget",
						FATAL);
			}
			// handle Y-scatterer differently
			if (Scatterer[1] != 1) {
				// do this based on scatterer info
				//void my_zset
				//   (const int N, const void *alpha, void *X, const int incX);
				int i;
				printf("Nx=%d, Ny=%d, Nfz0=%d, Nz=%d\n", Nx, Ny, Nfz0, Nz);
				for (i = 0; i < Nx * Ny * Nfz0; i++) {
					my_zset(Nz * 2, &complexZero, ZeroStartPtr, 1);
					ZeroStartPtr += Nz * Nv;
				}
			}

		} else { //  (dim != 5)
			DataStartOffset = 0;
		}

		// OOPS! fz is now complex!
		//  need to normalize each depth individually !!!
		// call function to normalize target function depths
		NormalizeFz();

		MATRIX temp;
		temp = ConvertToComplex(Target_f);  // Here I am freeing Target_f also!, but I need to keep have fz point to Target_f,

		freeMatrix(Target_f);

		Target_f = temp;

	}
	// End do only once!!!

	// begin DO FOR EACH SIGMA_H

	// Get the xy-plane for the target function for sigma_h
	// Construct T in time space coordinates
	// Assume it is square and use the X-dimension because
	// the Y-dimension is roughly half because of Fourier Transform
	t = newMatrix(Nx, Nx);
	Float *p;

	x = (Float) -Nx / 2.0;
	y = (Float) -Nx / 2.0;

	Float C; // The constant C is taken so that the spatial integral of the target function is unity
	C = GaussianConstant();

	Float sum = 0.0;
	p = t.data;
	for (ix = 0; ix < Nx; ix++) {
		for (iy = 0; iy < Nx; iy++) {
			value = Gaussian(x + ix, y + iy); // includes C*exp()

			sum += value;
			if (NORM)
				*p++ = value;
			else
				*p++ = C * value * C;
// try calling some cblas function here to sum the matrix

		}
	}
	printf("sum of target 2D Gaussian is %e \n", sum);
	if (NORM) {
		MATRIX temp = multiplyScalar(t, 1.0/sum);
		freeMatrix(t);
		t = temp;
	}

	MATRIX tk;
	if (targetData.useFFT) {
		// Fourier transform T
		tk = fftw(MatrixToFitsObject(t));
	} else {
		tk = ConvertToComplex(t);
	}

	// multiply by a 1D function of the vertical coordinate, z
	// z0 is an optional parameter and is the depth where the target function is peaked.
	// * f(z; z0) delta(alpha, theta)

	// take care of z-dependence as needed.

	// 1. copy fz into T
	// 2. scale fz (svcal or sscal) by tk(x,y)
	// 3. c/zset  to zero all non-scatterers

	FloatComplex *tkp;
	tkp = tk.data; // this is transformed 2D Gaussian
	if ((tk.nRow != Nx) || (tk.nCol != Ny))
		reportError("Bad dimensions in target function", "constructTarget",
				FATAL);

	FloatComplex *tp;
	tp = T.data + DataStartOffset; // position in Target of scatterer (vx,vy, or vz)
	// remember to account for scatterer offset in T (use DataStartOffset set above)

	// for each element in the Gaussian 2D function (xy-plane)
	for (ix = 0; ix < Nx; ix++) {
		for (iy = 0; iy < Ny; iy++) {

			Buildup(ix, iy, &T, tkp, &tp);

			tkp += 1;
		} // end for ix
	} // for iy

	freeMatrix(t);
	if (targetData.useFFT) {
		// need fftw_free here
		fftwf_free(tk.data);
	} else {
		freeMatrix(tk);
	}
	return &T;
}

/**
 * Write target function to output FITS file. Note: the output filename is assumed to be set using
 * newTarget.
 * @param[in] sigmah Current value for sigmah parameter, where sigmah is the full-width at half-maximum (FWHM)
 * of the Gaussian.
 * @param[in] sigmaStep The nth parameter value for sigmah.
 * @ingroup Target
 */
void writeTarget(Float sigmah, int sigmaStep) {
	// write the target function
	int NAlpha, NFz0;
	NAlpha = Target->nV;
	NFz0 = Target->nT;
	char **filenames;

	//  we want three target functions (vx,vy,vz) for each depth
	// Ok, I have to allocate a filename bank, based on Nfz0 and Nv.
	// Construct the filenames
	// Call writeMapDepthMatrix
	// Keep the filename bank around for the output files.

	SetupFileNameBank(targetData.outputBaseDir, NFz0, NAlpha);

	filenames = constructOutputFilenames(targetData.outputBaseDir, Constants.TargetFilename, 0, sigmaStep);
	FITSParameters headerParms;
	headerParms.n = 1;
	headerParms.comments[0] = Constants.ModelComment;
	headerParms.names[0] = Constants.ModelSigmaName;

	snprintf(&headerParms.scientificNotation[0][0], MAX_FITS_PARM_VALUE, FITSParametersConstants.sciFiFormat, sigmah);
	writeMapDepthMatrix(filenames, NAlpha*NFz0, *Target, targetData.useFFT, headerParms);
}

/**
 * Change the target function by recalculating with sigmah.
 * @param[in] sigmah Full-width at half maximum of function.
 * @param[in] sigmaStep The nth sigma step in the parameter range.
 * @ingroup Target
 */
void resetTarget(Float sigmah, int sigmaStep) {
	Float TwoSqrt2ln2 = 2.0 * sqrt(2.0 * log(2.0));

	Float sigma = sigmah / TwoSqrt2ln2;
	newGaussian(0.0, sigma);

	MAP_DEPTH_MATRIX *T;
	T = constructTarget(NormalizeFz, Buildup);

	Target = T;
    writeTarget(sigmah, sigmaStep);
}

void resetTargetLSQR(Float sigmah, int sigmaStep) {
	Float TwoSqrt2ln2 = 2.0 * sqrt(2.0 * log(2.0));

	Float sigma = sigmah / TwoSqrt2ln2;
	newGaussian(0.0, sigma);

	// Alright, Target_f needs to be oriented (1 x Nz) since
	// we are using constructTarget and it handles
	// multiple target depths, but we don't want multiple
	// target depths for LSQR, so LSQR will have
	// its own normalization and scaling (buildup).

	MAP_DEPTH_MATRIX *T;
	T = constructTarget(NormalizeFzLSQR, BuildupLSQR);
	Target = T;
    writeTarget(sigmah, sigmaStep);
}

/**
 * readFullTarget
 * Read the full target function from a fits file, not just the depth
 * function, f(z,z0).
 * @param[in] inputFilename Full path the to fits file containing the
 * target function.
 * @param[in] targetInfo
 */
/**
 * Read the two types of input data for a target function.
 * First type is a matrix (Nfz0 x Nz) containing
 * the z-dimension for possibly multiple target depth functions, f(z;z_0).
 * The second type is a complete target function of dimension
 * (Nx x Ny x Nz x Nbeta).
 * Multiple target functions are only for OLA inversions.
 * @param[in] inputFilename Character string for name of file containing target function.
 * @param[in] targetInfo Internal structure (MultiDimFits) that contains
 * information about an open fits file.
 * @param[in] N Size of Nx dimension of kernels
 * @param[in] Ny Size of Ny dimension of kernels
 * @param[in] Nz Size of Nz dimension of kernels
 * @param[in] sigmaCurrentValue Current value of sigma_h parameter (FWHM).
 * @param[in] sigmaCurrentStep sigmaStep The nth sigma step in the parameter range.
 * @param[in] baseDir pointer to BaseDirectory containing
 * the base directory for all output from the inversion.
 * @param[in] debug  When set to true, intermediate values are written to output directory.
 * @param[in] useFourierTransform true if Fourier transforming the target function on input.
 * @ingroup Target
 */
MATRIX readFullTarget(char *inputFilename, MultiDimFits targetInfo, int N,
		int Ny, int Nz, Float sigmaCurrentValue, Float sigmaCurrentStep,
		BaseDirectory *baseDir, bool debug, bool useFourierTransform) {
	MATRIX Targetf;
	int dim;
	dim = getProblemDimension();

	bool ok = false;
	printf("targetInfo:\n");
	printf("Nx=%d Ny= %d Nz= %d Nt= %d\n", targetInfo.max.x, targetInfo.max.y,
			targetInfo.max.z, targetInfo.max.t);
	printf("kernelInfo:\n");
	printf("Nx=%d Ny= %d Nz= %d \n", N, Ny, Nz);
	int NyTargetk;   // Ny for target in wave space
	NyTargetk = targetInfo.max.y / 2 + 1;
	if ((targetInfo.max.x == N) && (NyTargetk == Ny) && (targetInfo.max.z == Nz)
			&& (targetInfo.max.t == 1) && (dim == PDC.D3)) {
		ok = true;
	}
	if ((targetInfo.max.x == N) && (NyTargetk == Ny) && (targetInfo.max.z == Nz)
			&& (dim == PDC.D5) && (targetInfo.max.t == MAX_ALPHA)) {
		ok = true;
	}
	if (ok) {
		static MAP_DEPTH_MATRIX TargetTemp;

		newTarget(newMatrix(0, 0), N, Ny, Nz, baseDir, debug,
				useFourierTransform);
		TargetTemp = readMapDepthMatrixForTarget(inputFilename,
				useFourierTransform);

		Target = &TargetTemp;
		// Do a little 'fix up' of the target dimensions; since readMapDepthMatrixForTarget
		// puts the scatterer (Nbeta) dimension in nT, we need to move it to nV.
		// The nT (map) dimension for a target function is always 1, because the targets
		// are independent of maps.
		Target->nV = Target->nT;
		Target->nT = 1;

		// set sigma parameter steps to 1; somehow, check this and send message
		setSigmaStep(1);
		reportError(OLABaseSettingSigmah1, "readTarget", INFO);
		writeTarget(sigmaCurrentValue, sigmaCurrentStep);
	} else {
		char msg2[MAX_ERROR_MSG_LENGTH];
		if (dim == PDC.D3) {
			snprintf(msg2, MAX_ERROR_MSG_LENGTH,
					"Bad dimensions on target function. Expected (Nx,Ny,Nz) (%1d,%1d,%1d); Found (%1d,%1d,%1d)",
					N, (Ny - 1) * 2, Nz, targetInfo.max.x, targetInfo.max.y,
					targetInfo.max.z);
		} else {
			snprintf(msg2, MAX_ERROR_MSG_LENGTH,
					"Bad dimensions on target function. Expected (Nx,Ny,Nz,Nbeta) (%1d,%1d,%1d,%1d); Found (%1d,%1d,%1d,%1d)",
					N, (Ny - 1) * 2, Nz, MAX_ALPHA, targetInfo.max.x,
					targetInfo.max.y, targetInfo.max.z, targetInfo.max.t);
		}
		reportError(msg2, "readTarget", FATAL);
	}
	return Targetf;
}


MATRIX readZComponentTarget(char **inputFilenames, int nFilenames, int Nz) {
	MATRIX Targetf;

	// if there is only one filename, then let it be a matrix
	if (nFilenames == 1) {
		Targetf = toMatrix(readfits(inputFilenames[0]));
	} else {
		// allocate space for building the target matrix
		Targetf = newMatrixG(nFilenames, Nz, REAL);
		Float *targetFPtr;
		int n;

		targetFPtr = Targetf.data;
		for (n = 0; n < nFilenames; n++) {
			// read the fits file
			// check for compatible dimensions
			// if not correct size, report error and quit?
			// fill the correct row in the accumulation matrix

			// we want the number of rows to be the number of depths
			// and the number of columns to be Nz.  So, the function spans a row.
			VECTOR v;
			v = toVector(readfits(inputFilenames[n]));
			// Need to check the dimensions before inserting? What if it's not a vector?
			if (v.data == NULL)
				reportError("Encountered problem with target depth function, exiting", "readZComponentTarget", FATAL);
			insertVector(targetFPtr, v);
			freeVector(v);
			// advance to the next row in the target function
			targetFPtr += Nz;
		}
	}

	// we want the number of rows to be the number of depths
	// and the number of columns to be Nz.  So, the function spans a row.
	checkTargetDim(&Targetf, Nz);
	return Targetf;

}

/**
 * Read the two types of input data for a target function.
 * First type is a matrix (Nfz0 x Nz) containing
 * the z-dimension for possibly multiple target depth functions, f(z;z_0).
 * The second type is a complete target function of dimension
 * (Nx x Ny x Nz x Nbeta).
 * Multiple target functions are only for OLA inversions.
 * @param[in] targetInfo Internal structure containing info about the target function.
 * @param[in] N Size of Nx dimension of kernels
 * @param[in] Ny Size of Ny dimension of kernels
 * @param[in] Nz Size of Nz dimension of kernels
 * @param[in] sigmaCurrentValue Current value of sigma_h parameter (FWHM).
 * @param[in] sigmaCurrentStep sigmaStep The nth sigma step in the parameter range.
 * @param[in] baseDir pointer to BaseDirectory containing
 * the base directory for all output from the inversion.
 * @param[in] debug  When set to true, intermediate values are written to output directory.
 * @param[in] useFourierTransform true if Fourier transforming the target function on input.
 * @ingroup Target
 */
//readTarget(inputParms.targetInfo[], N, Ny, Nz, Nfz0, getSigma()->currentValue, getSigma()->currentStep,
//		inversionData->outputBaseDir, inversionData->debug, true);
// 1) send array and number of target depths
// 2) keep readTarget the same, call it, then move the info to the correct position in the nrhs
//    MAP_DEPTH_MATRIX
// Handle multiple target functions only for OLA inversions.
//
// If there is one file, it can have multiple dimensions and
// if nFilenames > 1, then enter a loop and build up the matrix.
// An error occurs if there are multiple filenames AND a file is NOT a vector.
// ---------------------------
// for each target depth file
//   read f(z;z_0) from file
//   construct 3D target function,
//   put it into the MAP_DEPTH_MATRIX	 for targets
// end for

// 1) send array and number of target depths
// 2) keep readTarget the same, call it, then move the info to the correct position in the nrhs
//    MAP_DEPTH_MATRIX

//
void readTarget(TargetInfo targetInfo, int N, int Ny, int Nz, Float sigmaCurrentValue,
 Float sigmaCurrentStep, BaseDirectory *baseDir, bool debug,
 bool useFourierTransform) {
//
// If there is one file, it could have multiple dimensions and
// if nFilenames > 1, then enter a loop and build up the matrix.
// An error occurs if there are multiple filenames AND a file is NOT a vector.
// ---------------------------
// for each target depth file
//   read f(z;z_0) from file
//   construct 3D target function,
//   put it into the MAP_DEPTH_MATRIX	 for targets
// end for


	char **inputFilenames;
	int nFilenames;

	inputFilenames = targetInfo.TargetFilenames;
	nFilenames = targetInfo.nTargetFiles;

	MATRIX Targetf;
	int dim;
	dim = getProblemDimension();
	if (dim > PDC.D2) {

		if (inputFilenames[0] != NULL) {
			// only if z > 1; i.e. 2d problem
			// read the header and determine the dimension 2D or 3D/4D
			MultiDimFits targetInfo;
			targetInfo = openMultiDimFitsRead(inputFilenames[0]);
			closeMultiDimFits(targetInfo);
			// if depth only provided
			if (targetInfo.dimension == 2) {
				// if 2D
				// Targetf can have multiple dimensions instead of 1 x Nz,
				// it would be Nfz0 x Nz.  Just insert a loop here
				// -- OR -- the inputFile could have multiple dimensions instead of 1 x Nz,
				// it would be Nfz0 x Nz, where each row contains a function at a different
				// depth.  Both are allowed.

				Targetf = readZComponentTarget(inputFilenames, nFilenames, Nz);

				// we want the number of rows to be the number of depths
				// and the number of columns to be Nz.  So, the function spans a row.
				checkTargetDim(&Targetf, Nz);

			} else {  // the entire function is provided
				// else 3D/4D
				if (nFilenames > 1)
					reportError("Only using first target function; multiple functions not implemented", "readTarget", WARNING);
				Targetf = readFullTarget(inputFilenames[0], targetInfo,
						 N, Ny, Nz, sigmaCurrentValue,
						 sigmaCurrentStep, baseDir, debug,
						 useFourierTransform);

			}
		} else {
			reportError(OLABaseMissing1dTargetFunc, "readTarget", FATAL);
		}
	} else {
		Targetf = newMatrix(1,1);
		Float *p;
		p = Targetf.data;
		p[0] = 1.0;
	}
	// initialize the target function
	// use Fourier transform = true
	if (Target == NULL) {
		newTarget(Targetf, N, Ny, Nz, baseDir, debug, useFourierTransform);
		resetTarget(sigmaCurrentValue, sigmaCurrentStep);

	}
}

/**
 * Returns a pointer to target function represented by a Map-Depth matrix.
 * @returns Pointer to MAP_DEPTH_MATRIX representing the generated target function.
 * @ingroup Target
 */
MAP_DEPTH_MATRIX *getTarget() {
	return Target;
}

/**
 * Free space associated with target function.
 * @ingroup Target
 */
void freeTarget() {
	freeMatrix(Target_f);
	destroyMapDepthMatrix(*Target);
}

