/*
 * main.c
 *
 *  Created on: Aug 16, 2010
 *      Author: Brenda Javornik (e-mail: brendaj (at) nwra.com)
 */

#include <string.h>
#include <math.h>
#include "Constants.h"
#include "Gaussian.h"
#include "Target.h"
#include "errorHandler.h"
#include "InversionErrorMessages.h"
#include "OLAErrorMessages.h"
#include "FileMapAndParameters.h"
#include "ProblemDimension.h"
#include "utilities.h"
#include "matrix.h"
#include "MapDepthMatrix.h"
#include "OLAInvert.h"
#include "Model.h"
#include "Scatterer.h"
#include "InversionData.h"
#include "ParameterLoop.h"
#include "Parameter.h"
#include "CrossTalk.h"

#define K2PICC_FLY 0

// M = # of travel time measurements = number of kernel-map pairs = t dimension in map-depth matrix
// P = # of perturbations = 1
// N = # of points in each x and y direction; Kernel files are N x N x z

/**
 * Check the dimensions of all the input matrices.
 * @param[in] K Kernel in frequency space
 * @param[in] Lambda Lambda Travel-time noise covariance matrix
 * @param[in] Theta Overlap matrix, describing expansion in terms of basis functions.
 * @param[in] tau Travel-time matrix
 * @param[in] FourierTransformed True if the Kernel data has been transformed.
 * @param[in] noMaps Boolean set to true if running inversion without maps; false otherwise.
 */
bool DimensionsOk(MAP_DEPTH_MATRIX K, MAP_DEPTH_MATRIX Lambda, MATRIX Theta,
	MAP_DEPTH_MATRIX tau,  bool FourierTransformed, bool noMaps) {
	bool ok = true;
	// check matrix dimensions for compatibility

	int Nx, Ny, M, Z;
	char msg2[MAX_ERROR_MSG_LENGTH];


	// pull dimensions from Kernel matrix
	Nx = K.nX;
	Z = K.nZ * K.nV; // add multiplier for 3d_vector K.nZ;
	M = K.nT;
	if (FourierTransformed) {
		Ny = Nx/2 + 1;
	} else {
		Ny = Nx;
	}

	// Make sure K.x = K.y = N
	if (K.nY != Ny) {
		ok = false;
		// error, Kernel files must have equal x and y dimensions
		reportError(OLAMainKernelSquare, "main", ERROR);
	}

	if (!noMaps) {
		// Make sure tau.x = tau.y = N is square.
		if ((tau.nX != Nx) || (tau.nY != Ny)) {
			ok = false;
			// error, map data must have equal x and y dimensions and be the same as Kernel files
			reportError(OLAMainMapSquare, "main", ERROR);
		}

		// Make sure tau.nT = K.nT = M; tau has the same number of maps as K.
		if (tau.nT != M) {
			ok = false;
			// error, number of map files must equal number of kernel files.
			reportError(OLAMainNMapNKernel, "main", ERROR);
		}
	}

	// Make sure Lambda is NxNxMxM
	// Make sure Lambda.x = Lambda.y = N
	// Make sure Lambda.nT = Lambda.nZ = M
	if ((Lambda.nX != Nx) || (Lambda.nY != Ny) || (Lambda.nT != M) || (Lambda.nZ != M)) {
		ok = false;
		// error, "Lambda must have dimensions %d x %d x %d x %d", N, N, M, M
		snprintf(msg2, MAX_ERROR_MSG_LENGTH, "%s %d x %d x %d x %d", OLAMainLambdaDim, Nx, Nx, M, M);
		reportError(msg2, "main", ERROR);
	}

	// Make sure Theta is ZxZ
	if ((Theta.nRow != Z) || (Theta.nCol != Z)) {
		ok = false;
		// error, "Theta must have dimensions %d x %d", Z, Z
		snprintf(msg2, MAX_ERROR_MSG_LENGTH, "%s %d x %d", OLAMainThetaDim, Z, Z);
		reportError(msg2, "main", ERROR);
	}
	int ZCompare = Z;
	if (getProblemDimension() == PDC.D5) {
		// Z can be Z/3
		ZCompare = ZCompare / MAX_ALPHA;
	}

	return ok;
}

bool DimensionsOkPrelim(MAP_DEPTH_MATRIX K, MAP_DEPTH_MATRIX Lambda, MATRIX Theta,
	MAP_DEPTH_MATRIX tau,  bool FourierTransformed, int numberOfMaps, bool noMaps) {
	bool ok = true;
	// check matrix dimensions for compatibility

	int Nx, Ny, M, Z;
	char msg2[MAX_ERROR_MSG_LENGTH];


	// pull dimensions from Kernel matrix
	Nx = K.nX;
	Z = K.nZ * K.nV; // add multiplier for 3d_vector K.nZ;
	M = numberOfMaps; // K.nT;
	if (FourierTransformed) {
		Ny = Nx/2 + 1;
	} else {
		Ny = Nx;
	}

	// Make sure K.x = K.y = N
	if (K.nY != Ny) {
		ok = false;
		// error, Kernel files must have equal x and y dimensions
		reportError(OLAMainKernelSquare, "main", ERROR);
	}

	if (!noMaps) {
		// Make sure tau.x = tau.y = N is square.
		if ((tau.nX != Nx) || (tau.nY != Ny)) {
			ok = false;
			// error, map data must have equal x and y dimensions and be the same as Kernel files
			reportError(OLAMainMapSquare, "main", ERROR);
		}
	}

	// Make sure Lambda is NxNxMxM
	// Make sure Lambda.x = Lambda.y = N
	// Make sure Lambda.nT = Lambda.nZ = M
	if ((Lambda.nX != Nx) || (Lambda.nY != Ny) || (Lambda.nT != M) || (Lambda.nZ != M)) {
		ok = false;
		// error, "Lambda must have dimensions %d x %d x %d x %d", N, N, M, M
		snprintf(msg2, MAX_ERROR_MSG_LENGTH, "%s %d x %d x %d x %d", OLAMainLambdaDim, Nx, Nx, M, M);
		reportError(msg2, "main", ERROR);
	}

	// Make sure Theta is ZxZ
	if ((Theta.nRow != Z) || (Theta.nCol != Z)) {
		ok = false;
		// error, "Theta must have dimensions %d x %d", Z, Z
		snprintf(msg2, MAX_ERROR_MSG_LENGTH, "%s %d x %d", OLAMainThetaDim, Z, Z);
		reportError(msg2, "main", ERROR);
	}
	int ZCompare = Z;
	if (getProblemDimension() == PDC.D5) {
		// Z can be Z/3
		ZCompare = ZCompare / MAX_ALPHA;
	}

	return ok;
}


/** Fill InversionData structure with data from input files.
 * @param[in] inputParms InputParms structure filled with filenames, directories and parameters
 * extracted from input file.
 * @param[in] commandLineOutputDir character string with new output directory specified on command line.
 * @param[in] debug	Boolean flag for writing intermediate matrix values.
 * @return InversionData structure filled with input data.
 */
InversionData *OLAInit(InputParms inputParms, char *commandLineOutputDir, bool debug) {

	printf("Memory Usage at Initialization: %ld\n", getMemoryReport());

	int dim;

	InversionData *inversionData;

	inversionData = (InversionData *) td_malloc(sizeof(InversionData));
	inversionData->debug = debug;

	// set problem dimension, which tells us how to interpret the
	// file map and parameters
	dim = getProblemDimension();
	if (commandLineOutputDir == NULL) {
		inversionData->outputBaseDir = newBaseDirectoryNrhs(inputParms.outputBaseDirName);
	} else {
		inversionData->outputBaseDir = newBaseDirectoryNrhs(commandLineOutputDir);
	}

	if (dim == PDC.D5) {
		// set the scatterer of interest
		if (!setScatterer(inputParms.scatterer))
			reportErrorInt(MainBadScatterNumber, inputParms.scatterer, "main", FATAL);
		printf("Solving for alpha = %c\n", translateScatterer(inputParms.scatterer));
	} else {
		setScatterer(0);
	}

	// read input files, K, t, Lambda; need read and write fitsio as a library.
	// 3d problem, read, transform and coalesce kernel files into giant matrix MFull
	bool useFourierTransform = true;
	printf("Reading first Kernel file ...\n");
	MAP_DEPTH_MATRIX K1 = readMapDepthMatrixKernel(inputParms.d3.InputKernelBaseDirName, inputParms.d3.KernelFilenames,
			1, useFourierTransform);

	MAP_DEPTH_MATRIX tau1;
	if (!inputParms.noMaps) {
		// Read, transform and coalesce map files into giant matrix dFull
		printf("Reading first travel time file ...\n");
		tau1 = readMapDepthMatrix(inputParms.d3.InputMapBaseDirName, inputParms.d3.MapFilenames,
				1, useFourierTransform);
	}
	// Read and transform noise covariance files into giant matrix Lambda
	printf("Reading noise covariance files ...\n");
	MAP_DEPTH_MATRIX Lambda;
	Lambda = readMapDepthMatrixFromOneFile(inputParms.InputICovarFileName, useFourierTransform);

	if ((dim == PDC.D2) && (K1.nZ > 1)) {
		//dim = 3;
		if (getProblemDimension() != dim)
			reportError(BadZDimForKernel, "OLAInversion.main", FATAL);
	}

	// handle Theta for 3D and 2D problems.  If 2D, set Theta to 1
	MATRIX Theta;
	if (dim > PDC.D2 ) {
		// Read Theta; overlap matrix
		if (inputParms.InputRegFileName != NULL)
			Theta = toMatrix(readfits(inputParms.InputRegFileName));
		else
			reportError(MissingThetaFile, "OLAInversion.main", FATAL);
	} else {
		Theta = newMatrix(1,1);
		Float *p;
		p = Theta.data;
		p[0] = 1.0;
	}

	// convert Theta to complex for homogeneous matrix multiplication
	MATRIX temp = ConvertToComplex(Theta);
	freeMatrix(Theta);
	Theta = temp;

	// check matrix dimensions for compatibility
	if (!DimensionsOkPrelim(K1, Lambda, Theta, tau1, useFourierTransform, inputParms.d3.nKernelMapFiles,
			inputParms.noMaps))
			reportError(IncompatibleMatrices, "OLAInit", FATAL);

	int N;				// Number of horizontal position; x and y dimension of Kernel
	int Ny;				// y dimension after FFT
	int M; 				// Number of measurements, denoted by a or c.

	N = K1.nX;
	Ny = K1.nY;

	int Nz = K1.nZ;

	destroyMapDepthMatrix(K1);
	if (!inputParms.noMaps)
		destroyMapDepthMatrix(tau1);

	//------- read all kernel and map files
	useFourierTransform = true;
	printf("Reading Kernel files ...\n");
	MAP_DEPTH_MATRIX K = readMapDepthMatrixKernel(inputParms.d3.InputKernelBaseDirName, inputParms.d3.KernelFilenames,
			inputParms.d3.nKernelMapFiles, useFourierTransform);
	MAP_DEPTH_MATRIX K2piCC;
	if (!K2PICC_FLY) {
		printf("Making copy of kernels ... \n");
		K2piCC = TdmCopy(K);
		printf("Copy complete.\n");
		TdmMultiplyScalarTakeConjugate(&K2piCC, (2.0*M_PI)*(2.0*M_PI));
	}

	MAP_DEPTH_MATRIX tau;
	if (!inputParms.noMaps) {
		// Read, transform and coalesce map files into giant matrix dFull
		printf("Reading travel time files ...\n");
		tau = readMapDepthMatrix(inputParms.d3.InputMapBaseDirName, inputParms.d3.MapFilenames,
				inputParms.d3.nKernelMapFiles, useFourierTransform);
	}
	//------- end read all kernel and map files


	// check matrix dimensions for compatibility
	if (!DimensionsOk(K, Lambda, Theta, tau, useFourierTransform, inputParms.noMaps))
			reportError(IncompatibleMatrices, "OLAInit", FATAL);

	inversionData->InputMatrices = newInputData();
	InputData *id;
	id = inversionData->InputMatrices;
	inversionData->IntermediateMatrices = (TempData *) td_malloc(sizeof(TempData));
	TempData *tempd;
	tempd = inversionData->IntermediateMatrices;
	id->K = K;
	id->C = Lambda;
	id->M = tau;
	id->Theta = Theta;
	if (!K2PICC_FLY)
		tempd->K2piCC = K2piCC;

	if (inputParms.lambdaSet2.min <= 0.0)
		reportError(BadSigmaMin, "OLAInit", FATAL);

	// Parameters is a pointer to the head of a linked list of parameters.
	// Mu, Nu, and Sigma are pointers to individual elements/nodes of the
	// parameter list.
	initLoopParameters(inputParms);

	printf("Reading target files ...\n");
	//read all target files and prepare target functions
	readTarget(inputParms.targetInfo, N, Ny, Nz, getSigma()->currentValue, getSigma()->currentStep,
			inversionData->outputBaseDir, inversionData->debug, true);

	M = K.nT;
	inversionData->outputModel = (Model *) td_malloc(sizeof(Model));
	// Hmm, need to know NFz0, but target has the info ... how to get it? Delay allocation of w,
	// until we are saving wk, then we'll  know Nfz0.
	initializeModel(inversionData->outputModel, N, Ny, M, Nz);
	if (inputParms.noMaps) {
		useNoMaps(inversionData->outputModel);
	}

	freeInputParms(inputParms);

	return(inversionData);
}
