/*
 * OLAInvert.c
 *
 *  Created on: Aug 16, 2010
 *      Author: Brenda Javornik (e-mail: brendaj (at) nwra.com)
 */

#include <string.h>
#include <math.h>
#include <complex.h>
#include "InvTypes.h"
#include "Model.h"
#include "errorHandler.h"
#include "OLAErrorMessages.h"
#include "matrix.h"
#include "MapDepthMatrix.h"
#include "linearEqSolve.h"
#include "ProblemDimension.h"
#include "Target.h"
#include "Gaussian.h"
#include "OneMatrix.h"
#include "Scatterer.h"
#include "Constants.h"
#include "InputData.h"
#include "InversionData.h"
#include "Parameter.h"
#include "CrossTalk.h"
#include "WeightSpread.h"

#define K2PICC_FLY 0

#ifdef SINGLE
#define cabs cabsf
#define creal crealf
#endif

/**
 * Construct a special matrix [A, C; C', 0]
 * @param[in] A Matrix
 * @param[in] C Matrix Constants, integral sum of kernel
 * @return special matrix as defined above
 */
MATRIX ConstructMatrixB(MATRIX A, MATRIX C) {
	MATRIX B;
	int nRows, nCols;
	MATRIX CT;

	if ((A.nRow != C.nRow))
		reportError(OLARowARowC, "ConstructMatrixB", FATAL);
	if (((getProblemDimension() != PDC.D5) && (C.nCol != 1)) || ((getProblemDimension() == PDC.D5) && (C.nCol != 3)))
		reportError(OLACWrongColumns, "ConstructMatrixB", FATAL);

	nRows = A.nRow;
	nCols = A.nCol;
	if ((C.nRow > 0) && (C.nCol > 0)) {
		// add C to the matrix
		nRows += C.nCol; // remember, we are appending transpose of C
						// to bottom of matrix B
		nCols += C.nCol;
		CT = transpose(C);
	}

	B = newComplexMatrix(nRows, nCols);
	// start copying memory
	FloatComplex *pto;
	FloatComplex *pB, *pA, *pC;
	pB = B.data;
	pA = A.data;
	pC = C.data;
	int r, c;
	for (r=0; r<A.nRow; r++) {
		// copy a row from A
		// memcpy(to, from, size)
		memcpy(pB, pA, (A.nCol * sizeof(FloatComplex)));
		pB += A.nCol; // move the pointer in B to the end of row from A
		// copy next row from C
		// memcpy(to, from, size)
		memcpy(pB, pC, (C.nCol * sizeof(FloatComplex)));
		pB += C.nCol; // move the pointer in B to the next row
		pA += A.nCol; // move pointer in A to next row
		pC += C.nCol;
	}
	// copy the rows from C'
	pC = CT.data;
	for (r=0; r<CT.nRow; r++) {
		// memcpy(to, from, size)
		// copy a row from C'
		memcpy(pB, pC, (CT.nCol * sizeof(FloatComplex)));
		// copy next row from C
		// insert zeros into remainder of row
		pto = pB + CT.nCol;
		for (c=0; c<C.nCol; c++) {
			*pto++ = 0.0;
		}
		pB += nCols; // move the pointer in B to the next row
		pC += CT.nCol;
	}

	if ((C.nRow > 0) && (C.nCol > 0)) {
		freeMatrix(CT);
	}
	return B;
}

/**
 * Construct a special matrix [w; Lambda]
 * @param[in] wt Matrix containing weight vector and other Lagrange parameters.
 * @return special matrix [w; Lambda]
 */
MATRIX ExtractMatrixW(MATRIX wt) {
	MATRIX w;
	int nRowsW;

	if (getProblemDimension() == PDC.D5)
		nRowsW = wt.nRow - 3;
	else
		nRowsW = wt.nRow - 1;
	w = newComplexMatrix(nRowsW, wt.nCol);
	FloatComplex *p, *q;
	p = wt.data;
	q = w.data;
	// memcpy(to, from, size)
	memcpy(q, p, (nRowsW * wt.nCol * sizeof(FloatComplex)));
	return w;
}

/**
 * Construct a special matrix [t; constant]
 * @param[in] t Matrix containing averaging kernel and target data
 * @return Matrix with special format.
 */
MATRIX ConstructMatrixT(MATRIX t) {
	MATRIX t2;
	int nRows, nCols;
	int nAlpha;
	int a;

	if (getProblemDimension() == PDC.D5) {
		nRows = t.nRow + 3;
		nAlpha = 3;
	} else {
		nRows = t.nRow + 1;
		nAlpha = 1;
	}
	nCols = t.nCol;

	Float *Scatterer = getScatterer();

	t2 = newComplexMatrix(nRows, nCols);
	// start copying memory
	FloatComplex *pt2, *pt;
	pt2 = t2.data;
	pt = t.data;

	// remember, t and t2 are row-major and
	// we are adding three additional rows at the bottom

	// copy all of t into t2
	// memcpy(to, from, size)
	memcpy(pt2, pt, (t.nCol * t.nRow * sizeof(FloatComplex)));
	pt2 += t.nCol * t.nRow; // move the pointer in t2 to the end of t data

		// put additional rows of constants
		for (a=0; a<nAlpha; a++) {
			// Here, set to 1.0 for the scatterer of interest ....
			// NOTE:  if dim == 2 or 3, then set Scatterer[0] = 1 because when constructing
			// tm matrix, Scatterer[0] is inserted into tm, and it must be one.
			//*pt2 = 1.0 / (TWO_PI_SQUARED);
			int c;
			// for each column of t
			for (c=0; c<t.nCol; c++) {
				*pt2 = Scatterer[a] / (TWO_PI_SQUARED);
				pt2 += 1;  // move the pointer to the next element
			}
		}
	return t2;
}

/**
 * Perform inversion using OLA approach.
 * @param[in] K Kernel in frequency space
 * @param[in] K2piCC The complex conjugate of K in frequency space multiplied by (2pi)^2;
 * @param[in] Lambda Lambda Travel-time noise covariance matrix
 * @param[in] Theta Overlap matrix, describing expansion in terms of basis functions.
 * @param[in] T Target function described in matrix form.
 * @param[in,out] model A Model Object.  Inversion results are returned in this object.
 * @param[in] debugFilename BaseDirectory structure to manage output of debug file names.
 * @param[in] debug bool flag; true if debug; false otherwise.
 * Precalculate the Target function for x, y and z values, then Fourier transform it.
 */
void OLAInvert(MAP_DEPTH_MATRIX K, MAP_DEPTH_MATRIX K2piCC, MAP_DEPTH_MATRIX Lambda,
		 MATRIX Theta,
		 MAP_DEPTH_MATRIX T, Model *model, BaseDirectory *debugFilename, bool debug) {

	bool kEqual0;
	int N;				// Number of horizontal position; x and y dimension of Kernel
	int Ny;				// y dimension after FFT
	MATRIX C;
	MATRIX Ak;
	MATRIX td;
	MATRIX tdK2piCC;
	MATRIX LambdaTd;
	MATRIX *One;
	Float hk;
	bool dumpDebug = false;
	bool debugPosition = true;

	N = K.nX;
	Ny = K.nY;

	int Nz = K.nZ;


	// See if Theta is the identity matrix.
	// Only do this once (so make it static).
	static bool thetaIdentityDetermined = false;
	static bool thetaIdentity = false;
	if (!thetaIdentityDetermined) {
		thetaIdentity = isIdentity(Theta);
		if (thetaIdentity)
			printf("Theta is the identity matrix\n");
		thetaIdentityDetermined = true;
	}

	hk = (2.0 * M_PI) / N;

	int dim = getProblemDimension();
	// create and set matrix to 1
	One = GetOne(Nz);

	int kx, ky;

#ifdef CALC_COND
	MATRIX condNums;
	Float *condp;
	MATRIX avgs;
	FloatComplex *avgsp;

	condNums = newMatrix(N,Ny);
	condp = condNums.data;
	avgs = newMatrixG(N,Ny,COMPLEX);
	avgsp = avgs.data;
#endif

	//for each wavenumber, k = (kx, ky)
	for (kx = 0; kx < N; kx++) {
		for (ky = 0; ky < Ny; ky++) {
			kEqual0 = false;
			if ((kx < 3) && (ky < 3) && debug) {
				dumpDebug = true;
				addParmsToFilename(debugFilename, kx, ky, debugPosition);
			} else {
				dumpDebug = false;
			}

			// Calculate A in frequency space
			//Ak = CalculateAk(c, a, k);
			// K is already multiplied by (2pi)^2
			//value = value * conjugate(Kc) * Ka + mu * Lambda(k);

			//td = extract(K(k))
			td = extractMapDepth(kx, ky, K);
			if (!K2PICC_FLY)
				tdK2piCC = extractMapDepth(kx, ky, K2piCC);

			LambdaTd = extractMapDepth(kx, ky, Lambda);

			// A(k) = K*(k) K(k)' + Nz* mu * Lambda(k)
			// transpose of K(k) taken in multiplyMMplusCZ routine

			// if crosstalk included multiply K* by nu for beta != alpha,
			// i.e. the flows for which we are NOT solving.

			MATRIX KTheta;
			if (dim >= 3) {
				if (K2PICC_FLY) {
					MATRIX ttd = transpose(td);
					KTheta = multiplyMMFull('C', 'N', (2.0*M_PI)*(2.0*M_PI), 0.0, ttd, Theta);
					freeMatrix(ttd);
				} else {
					if (thetaIdentity)
						KTheta = tdK2piCC;
					else
						KTheta = multiplyMM2('N', 'N', tdK2piCC, Theta);
				}
			} else {

				/*
				 *
				tdK2piCC = (2.0*M_PI)*(2.0*M_PI) td
						MATRIX multiplyScalar(MATRIX m, Float scalar)
						MATRIX conjugate(MATRIX m)
						 */
				if (K2PICC_FLY) {
					MATRIX temp;
					temp = multiplyScalar(td, (2.0*M_PI)*(2.0*M_PI));
					KTheta = conjugate(temp);
					freeMatrix(temp);
				} else {
					KTheta = tdK2piCC;
				}
			}

			if (getCrossTalk()) {
				MATRIX td2;
				Float valueXT;
				valueXT = getNu()->currentValue;
				td2 = multipyForCrosstalk(td, valueXT, Nz);
				Ak = multiplyMMplusCZ(KTheta, td2, model->mu, LambdaTd);
				freeMatrix(td2);
			} else {
				Ak = multiplyMMplusCZ(KTheta, td, model->mu, LambdaTd);
			}

			// Calculate C
			// Dimensions of C are M x 1
			// Because the Fourier transform of the kernel contains
			// the sum of the elements at k=0; multiply by 2pi^2 to
			// remove scaled or normalization during transform.

			// Because the Fourier transform of the kernel contains
			// the sum of the elements at k=0; multiply by 2pi^2 to
			// remove scaled or normalization during transform.
			//MATRIX td2pi = multiplyScalar(td, TWO_PI_SQUARED);
			// for (k == 0) do something special
			if ((kx == 0) && (ky == 0)) {
				// assume K(kx=0,ky=0) is the sum of the magnitude of the real space K matrix
				kEqual0 = true;
				if (dim >= 3) {

					MATRIX temp;
					//MATRIX multiplyMMFull(char TransA, char TransB, Float alpha, Float beta, MATRIX A, MATRIX B)
					// C = alpha A x B + beta C
					temp = multiplyMMFull('N', 'N', TWO_PI_SQUARED, 0.0, td, Theta);
					C = multiplyMM2('N', 'N', temp, *One);
					freeMatrix(temp);
				} else {
					C = multiplyScalar(td, TWO_PI_SQUARED);
				}
			} else {
				// Set C = NullMatrix;
				C.nCol = 0;
				C.nRow = 0;
			}

			if (getWeightSpread()) {
				Float value;
				value = getEpsilon()->currentValue;
				value = value * (1.0/TWO_PI_SQUARED);     // equivalent to 1/(N^2 * hk^2)
				addForWeightSpread(&Ak, value);
			}

			// Calculate t
			MATRIX Tk;
			Tk = extractMapDepth(kx, ky, T);

			// For the Kernel, we have already extracted this for the A matrix calculation;
			MATRIX t;
			// t should have dimensions: M x 1
			if (kEqual0) {
				t = multiplyMMFull('N', 'T', hk*hk, 0.0, KTheta, Tk);
			} else {
				t = multiplyMM2('N', 'T', KTheta, Tk);
			}

			if (kEqual0) {
				MATRIX temp = multiplyScalar(Ak, hk*hk*TWO_PI_SQUARED); // equivalent to (hk)^4 N^2
				freeMatrix(Ak);
				Ak = temp;
			} else {
				MATRIX temp = multiplyScalar(Ak, TWO_PI_SQUARED); // equivalent to (hk)^2 N^2
				freeMatrix(Ak);
				Ak = temp;
			}

#ifdef CALC_COND
				*condp++ = calcCondNum(Ak);
#endif

			MATRIX w;
			// Construct Bw = t
			if (kEqual0) {
				MATRIX B = ConstructMatrixB(Ak, C);
				MATRIX tm = ConstructMatrixT(t);
#if SVD
				w = linearEqSolveSvd(B, tm);
#else
				// for LUDecomp and Cholesky, use LUDecomp because, B is NOT
				// positive definite because of the zeros in the bottom right-hand corner
				w = linearEqSolve(B, tm);
#endif
				// check for error in linear eq solver
				FloatComplex *p;
				p = w.data;

				if (isnan(creal(p[0]))) {
					reportError("Matrix [A C; C^T 0] cannot be inverted", "OLAInvert", FATAL);
				}

				// extract w from result
				MATRIX temp = ExtractMatrixW(w);
				freeMatrix(w);
				w = temp;
				if (dumpDebug) {
					bool compress = false;
					writefits(finishFilename(debugFilename, "B", compress, debugPosition), MatrixToFits3dObject(B));
					writefits(finishFilename(debugFilename, "tm", compress, debugPosition), MatrixToFits3dObject(tm));
				}
				freeMatrix(B);
				freeMatrix(tm);
			} else {
				// linear Eq solve for w and lambda, if (k == 0)
				//  solve for w(k) if (k != 0)
				//--------------------
#ifdef CHOLESKY
				w = linearEqSolveCholesky(Ak, t);
#elif SVD
				w = linearEqSolveSvd(Ak, t);
#else
				w = linearEqSolve(Ak, t);
#endif

				// check for error in linear eq solver
				FloatComplex *p;
				p = w.data;

				if (isnan(creal(p[0]))) {
					reportError(OLAMatrixAInvert, "OLAInvert", FATAL);
				}
			}

			if (dumpDebug) {
				bool compress = false;
				writefits(finishFilename(debugFilename, "td", compress, debugPosition), MatrixToFits3dObject(td));
				if (!K2PICC_FLY)
					writefits(finishFilename(debugFilename, "tdK2piCC", compress, debugPosition), MatrixToFits3dObject(tdK2piCC));
				writefits(finishFilename(debugFilename, "KTheta", compress, debugPosition), MatrixToFits3dObject(KTheta));
				writefits(finishFilename(debugFilename, "LambdaTd", compress, debugPosition), MatrixToFits3dObject(LambdaTd));
				writefits(finishFilename(debugFilename, "Ak", compress, debugPosition), MatrixToFits3dObject(Ak));
				writefits(finishFilename(debugFilename, "t", compress, debugPosition), MatrixToFits3dObject(t));
				writefits(finishFilename(debugFilename, "w", compress, debugPosition), MatrixToFits3dObject(w));
			}

#ifdef CALC_COND
			    // calculate the mean of the weights
				int nmaps;
				FloatComplex *wp;
				wp = w.data;
				FloatComplex sum;
				sum = 0.0 + 0.0*I;
				for (nmaps=0; nmaps<w.nRow*w.nCol; nmaps++) {
					//mean of (w matrix)
					sum += cabs(*wp++);
				}
				// keep value in matrix mean
				*avgsp++ = sum/(w.nRow*w.nCol);
#endif


			// save w(k)
			saveW(model, w, (int) kx, (int) ky);

			// free some memory
			if (dim >= 3) {
				if (!thetaIdentity)
					freeMatrix(KTheta);
			}
			if (kEqual0) {
				freeMatrix(C);
			}
			freeMatrix(Ak);
			freeMatrix(t);
			freeMatrix(w);
		}
	}

#ifdef CALC_COND
		bool compress = false;
		if (!dumpDebug)
		  addParmsToFilename(debugFilename, 1,1, debugPosition);
		writefits2d(finishFilename(debugFilename, "condNum", compress, debugPosition), MatrixToFitsObject(condNums));
		fits3d avgsAsFits3d;
		avgsAsFits3d = MatrixToFits3dObject(avgs);
		writefits(finishFilename(debugFilename, "avgw", compress, debugPosition), avgsAsFits3d);
		freeFits3d(avgsAsFits3d);
        freeMatrix(condNums);
        freeMatrix(avgs);
#endif
}

/**
 * Setup call to OLA inversion by extracting the input matrices and intermediate
 * matrices.  Setup the output files with the parameter steps. Record
 * parameter values in model and inversion data structure.
 * @param[in,out] inversionData Pointer to structure containing all inversion data.
 */
void OLAInvert2(InversionData *inversionData) {
	InputData *id;
	id = inversionData->InputMatrices;
	TempData *tempd;
	tempd = inversionData->IntermediateMatrices;

	int parmSteps[NUM_PARAMETERS];  // {mu, sigma, nu}
	Float parmValues[NUM_PARAMETERS];  // {mu, sigma, nu}
	int nParms;

	doSomething(parmValues, parmSteps, &nParms);
	setParameters2(inversionData->outputModel, parmValues, nParms);

	bool debugPosition = false;
	addParmsToFilename2(inversionData->outputBaseDir, parmSteps, nParms, debugPosition);
	OLAInvert(id->K, tempd->K2piCC, id->C, id->Theta, *getTarget(), inversionData->outputModel,
			inversionData->outputBaseDir, inversionData->debug);
}
