/*
 * ParameterLoop.c
 *
 *  Created on: Sep 8, 2010
 *      Author: Brenda Javornik (e-mail: brendaj (at) nwra.com)
 */

#include <stdlib.h>
#include <math.h>
#include "Constants.h"
#include "utilities.h"
#include "FileMapAndParameters.h"
#include "matrix.h"
#include "MapDepthMatrix.h"
#include "Model.h"
#include "Target.h"
#include "Gaussian.h"
#include "OLAInvert.h"
#include "BaseDirectory.h"
#include "OneMatrix.h"
#include "lambda.h"
#include "ParameterLoop.h"
#include "linearEqSolve.h"
#include "Parameter.h"
#include "IgnoreK0.h"
#include "Scatterer.h"

/**
 * Loops over mu and sigma parameters and calls inversion function.
 * ParameterLoop .....
 * initializes model,
 * loops through sigma values,
 * 		loops through mu values,
 * 			performs inversion,
 * 			outputs model.
 */
/**
 * Advance to the next parameter values
 * @param[in] inversionData Pointer to data
 */
void OLANextParameter(InversionData *inversionData) {
    	getParameters()->increment(getParameters());
    	// no need to increment sigma; we only increment sigma
    	// when mu has exceeded its range.  Once mu has exceeded
    	// its range, mu, will signal sigma to increase.
}

/**
 * Check if more parameter values
 * @param[in] inversionData Pointer to data
 * @return TRUE if more parameters to use; false if all parameters have been used.
 */
bool OLAMoreParameterSteps(InversionData *inversionData) {
	return(getParameters()->more(getParameters()));
}


/**
 * Complete inversion calculation.
 * @param[in] inversionData Pointer to data:
 *  K 4D-matrix containing Kernel data in wavespace.
 *  K2piCC 4D-matrix containing K*(2pi)^2 time conjugate transpose of kernel data, K, in wavespace.
 *  Lambda 4D-matrix containing covariance data, in wavespace.
 *  Theta 2D-matrix containing overlap data for integrating over z-dimension.
 *  tau 4D-matrix containing travel-map data, in wavespace.
 *  Targetf 2D-matrix containing z-dimension of target function.
 *  baseDir BaseDirectory structure which manages the output filenames.
 *  muSet Range specifier for parameter, mu.
 *  sigmaSet Range specifier for parameter, sigma.
 *  debug Boolean flag: true if writing intermediate matrix values during inversion.
 */
void OLAComplete(InversionData *inversionData) {
	Model *model;
	BaseDirectory *baseDir;

	model = inversionData->outputModel;
	baseDir = inversionData->outputBaseDir;
	MAP_DEPTH_MATRIX K, tau, Lambda;
	K = inversionData->InputMatrices->K;
	tau = inversionData->InputMatrices->M;
	Lambda = inversionData->InputMatrices->C;

	calculateDerivedValues(model, K, tau, Lambda, false);

	// write results
	writeQ(model, baseDir);  // because q.fits is implied
	writeAvgKern(model, baseDir);
	writeNoise(model, baseDir);

	// persist the inversion for later use with different tau values.
	writeW(model, baseDir);  // because w.fits is implied
    // if solving for v_z then remove k=0
	if (solvingVZ()) {
		// now set w(k=0) = 0 and calculate derived values and write results
		ZeroK0(&(model->w));
		setIgnoreK0();
		calculateDerivedValues(model, K, tau, Lambda, false);

		// write results
		writeQ(model, baseDir);  // because q.fits is implied
		writeAvgKern(model, baseDir);
		writeNoise(model, baseDir);

		// persist the inversion for later use with different tau values.
		writeW(model, baseDir);  // because w.fits is implied
		clearIgnoreK0();
	}
}

/**
 * Free memory
 * @param[in] inversionData Pointer to data
 */
void OLADestroy(InversionData *inversionData) {

	linearEqSolveCleanup();

	// free some memory
	DestroyOne();

	InputData *inData = inversionData->InputMatrices;
	destroyInputData(inData);

	TempData *tempd = inversionData->IntermediateMatrices;
	destroyMapDepthMatrix(tempd->K2piCC);

	//destroyMapDepthMatrix(*getTarget());
	freeTarget();

	Model *model = inversionData->outputModel;
	freeModel(model);
	td_free(model, sizeof(Model));

	destroyBaseDirectory(inversionData->outputBaseDir);
	td_free(inData, sizeof(InputData));
	td_free(tempd, sizeof(TempData));
	td_free(inversionData, sizeof(InversionData));

	destroyParameters(getParameters());

	printf("Memory Usage  %ld \tFinal\n", getMemoryReport());

}
